#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include "defines.h"
#include "vpdefines.h"
#include "globals.h"
#include "globals_gtk.h" 

#include "protos_volume.h"
#include "protos_gui.h"


void load_volume(){
   unsigned char *density; /* buffer for density data */
   unsigned density_size;  /* size of density data */
   void *volume;	         /* volume data */
   unsigned volume_size;   /* size of volume */

   fprintf(stderr, "Loading volume...");
   
   switch(_PixelFormat){
   case C_INTENSITY8:
      density = _rawdata;
      bytesPerVoxel = sizeof(RawVoxel);
      normalOffset = vpFieldOffset(dummy_voxel, normal);
      normalSize = sizeof(short);
      normalMax = VP_NORM_MAX;
      densityOffset = vpFieldOffset(dummy_voxel, density);
      densitySize = sizeof(unsigned char);
      densityMax = 255;
      gradientOffset = vpFieldOffset(dummy_voxel, gradient);
      gradientSize = sizeof(unsigned char);
      gradientMax = VP_GRAD_MAX;
      break;
   case C_INTENSITY16:
      density = _rawsdata;
      bytesPerVoxel = sizeof(RawShortVoxel);
      normalOffset = vpFieldOffset(dummy_short_voxel, normal);
      normalSize = sizeof(short);
      normalMax = VP_NORM_MAX;
      densityOffset = vpFieldOffset(dummy_short_voxel, density);
      densitySize = sizeof(unsigned short);
      densityMax = 65535;
      gradientOffset = vpFieldOffset(dummy_short_voxel, gradient);
      gradientSize = sizeof(unsigned char);
      gradientMax = VP_GRAD_MAX;
      break;
   }

   /* create a context */
   vpc = vpCreateContext();

   /* describe the layout of the volume */
   vpSetVolumeSize(vpc, _VolumeDimensions.x, _VolumeDimensions.y, _VolumeDimensions.z);
   vpSetVoxelSize(vpc, bytesPerVoxel, VOXEL_FIELDS, SHADE_FIELDS, CLSFY_FIELDS);
   vpSetVoxelField(vpc, NORMAL_FIELD, normalSize, normalOffset, normalMax);
   vpSetVoxelField(vpc, DENSITY_FIELD, densitySize, densityOffset, densityMax);
   vpSetVoxelField(vpc, GRADIENT_FIELD, gradientSize, gradientOffset, gradientMax);

   /* allocate space for the volume */
   density_size = _VolumeDimensions.x * _VolumeDimensions.y * _VolumeDimensions.z * densitySize;
   volume_size =  _VolumeDimensions.x * _VolumeDimensions.y * _VolumeDimensions.z * bytesPerVoxel;
   volume = malloc(volume_size);
   if (density == NULL || volume == NULL) {
	   fprintf(stderr, "out of memory\n");
	   exit(-1);
   }
   vpSetRawVoxels(vpc, volume, volume_size, bytesPerVoxel,
		            _VolumeDimensions.x * bytesPerVoxel,
		            _VolumeDimensions.x * _VolumeDimensions.y * bytesPerVoxel);

   /* compute surface normals (for shading) and
      gradient magnitudes (for classification) */
   if (vpVolumeNormals(vpc, density, density_size, DENSITY_FIELD, GRADIENT_FIELD, NORMAL_FIELD) != VP_OK) {
	   fprintf(stderr, "VolPack error: %s\n",
	   vpGetErrorString(vpGetError(vpc)));
	   exit(1);
   }

   /* allocate memory for the image buffer */
   image = calloc(_VolumeDimensions.x * _VolumeDimensions.y * 3, 1);
   
   fprintf(stderr, "done\n");
}

void classify_volume(){
   float *density_ramp;	   /* opacity as a function of density */
   float *gradient_ramp;   /* opacity as a function of gradient magnitude */

   apply_classify_page();
   density_ramp = malloc(sizeof(float) * (densityMax + 1));
   gradient_ramp = malloc(sizeof(float) * (gradientMax + 1));

   fprintf(stderr, "Classifying volume...");

   /* set the classification function */
   vpRamp(density_ramp, sizeof(float), numDensity, DensityRampX, DensityRampY);
   vpSetClassifierTable(vpc, DENSITY_PARAM, DENSITY_FIELD, density_ramp, sizeof(float) * (densityMax + 1));
   vpRamp(gradient_ramp, sizeof(float), numGradient, GradientRampX, GradientRampY);
   vpSetClassifierTable(vpc, GRADIENT_PARAM, GRADIENT_FIELD, gradient_ramp, sizeof(float) * (gradientMax + 1));
   vpSetd(vpc, VP_MIN_VOXEL_OPACITY, 0.05);

   /* classify */
   if (vpClassifyVolume(vpc) != VP_OK) {
	    fprintf(stderr, "VolPack error: %s\n",
		 vpGetErrorString(vpGetError(vpc)));
	    exit(1);
	}

   free(density_ramp);
   free(gradient_ramp);

   fprintf(stderr, "done\n");
}

void shade_volume(){
   int firstTime = 0;

   if(shade_table == NULL || weight_table == NULL) firstTime = 1;
   free(shade_table);
   free(weight_table);
   shade_table = malloc(shade_table_size());
   weight_table = malloc(weight_table_size());
   
   /* set the shading parameters */ 
   vpSetLookupShader(vpc, 3, numMaterials, NORMAL_FIELD, shade_table, shade_table_size(), DENSITY_FIELD, weight_table,
                     weight_table_size());
   read_materials();
   set_materials();
   
   if(firstTime){
      read_lights();
      set_lights();
      vpSetDepthCueing(vpc, 1.4, 1.5);
      vpEnable(vpc, VP_DEPTH_CUE, 1); 
      
      /* set the initial viewing parameters */
      vpCurrentMatrix(vpc, VP_MODEL);
      vpSeti(vpc, VP_CONCAT_MODE, VP_CONCAT_LEFT);
      vpRotate(vpc, VP_Y_AXIS, 130.0);
      vpRotate(vpc, VP_X_AXIS, -15.0);
      vpCurrentMatrix(vpc, VP_PROJECT);
      vpIdentityMatrix(vpc);
      vpWindow(vpc, VP_PARALLEL, viewVolume[0][0], viewVolume[0][1], viewVolume[1][0], viewVolume[1][1], 
               viewVolume[2][0], viewVolume[2][1]);
      vpCurrentMatrix(vpc, VP_MODEL);
   }
   else{
      read_lights();
      set_lights();
   }

   /* compute weight lookup table */
   read_weights();
   set_weight_table();
      
   /* compute shading lookup table */
	if (vpShadeTable(vpc) != VP_OK) {
	   fprintf(stderr, "VolPack error: %s\n",
	   vpGetErrorString(vpGetError(vpc)));
	   exit(1);
	}
}

void render_volume(){
   
   /* set the image buffer */ 
   vpSetImage(vpc, image, _VolumeDimensions.x * imageScaleX, _VolumeDimensions.y * imageScaleY,
              3 * _VolumeDimensions.x * imageScaleX, VP_RGB);
   
   /* render and store the images */
   vpSetd(vpc, VP_MAX_RAY_OPACITY, 0.95);
   
   /* render */
   if (vpRenderClassifiedVolume(vpc) != VP_OK) {
	   fprintf(stderr, "VolPack error: %s\n",
		vpGetErrorString(vpGetError(vpc)));
		exit(1);
	}
   
   on_darea_expose(darea, NULL, NULL);

}

int shade_table_size(){
   return sizeof(float) * (normalMax + 1) * numMaterials * 3;
}

int weight_table_size(){
   return sizeof(float) * (densityMax + 1) * numMaterials;
}


void apply_classify_page(){
   int i, j;

   printf("numDensity = %d\n", numDensity);
   for(i = 0, j = 0; i < (numDensity / 2); i++){
      DensityRampX[j] = atoi( (char*) gtk_editable_get_chars(GTK_EDITABLE(dpx[i][0]), 0, -1));
      DensityRampY[j] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(dpy[i][0]), 0, -1));
      j++;
      DensityRampX[j] = atoi( (char*) gtk_editable_get_chars(GTK_EDITABLE(dpx[i][1]), 0, -1));
      DensityRampY[j] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(dpy[i][1]), 0, -1));
      j++;
      printf(" %d %d %.3f %.3f --", DensityRampX[j-2], DensityRampX[j-1], DensityRampY[j-2], DensityRampY[j-1]);
   }
   printf("\n numGradient = %d\n", numGradient);
   for(i = 0; i < numGradient; i++){
      GradientRampX[i] = atoi( (char*) gtk_editable_get_chars(GTK_EDITABLE(gpx[i]), 0, -1));
      GradientRampY[i] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(gpy[i]), 0, -1));
      printf(" %d %.3f --", GradientRampX[i], GradientRampY[i]);
   }
   printf("\n");
}

void set_weight_table(){
   int i, j;
   float *cursor;
   
   cursor = weight_table;
   for(i = 0; i < 256; i++){
      if(i >= weightTable[0][0] && i <= weightTable[0][1]){
         for(j = 0; j < numMaterials; j++){
            if(j == 0) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }
      }
      else if(i >= weightTable[1][0] && i <= weightTable[1][1]){
         for(j = 0; j < numMaterials; j++){
            if(j == 1) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }
      }
      else if(i >= weightTable[2][0] && i <= weightTable[2][1]){
         for(j = 0; j < numMaterials; j++){
            if(j == 2) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }
      }
      else if(i >= weightTable[3][0] && i <= weightTable[3][1]){
         for(j = 0; j < numMaterials; j++){
            if(j == 3) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }
      }
      else if(i >= weightTable[4][0] && i <= weightTable[4][1]){
         for(j = 0; j < numMaterials; j++){
            if(j == 4) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }
      }
      else{
         for(j = 0; j < numMaterials; j++){
            if(j == 5) *cursor = 1.0;
            else *cursor = 0.0;
            cursor++;
         }   
      }
         
   }  
}



void set_materials(){

   switch(numMaterials){
   case 6:
      vpSetMaterial(vpc, VP_MATERIAL5, VP_AMBIENT, VP_BOTH_SIDES, matAmb[5].red, matAmb[5].green, matAmb[5].blue);
      vpSetMaterial(vpc, VP_MATERIAL5, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[5].red, matDiff[5].green, matDiff[5].blue);
      vpSetMaterial(vpc, VP_MATERIAL5, VP_SPECULAR, VP_BOTH_SIDES, matSpec[5].red, matSpec[5].green, matSpec[5].blue);
      vpSetMaterial(vpc, VP_MATERIAL5, VP_SHINYNESS, VP_BOTH_SIDES, matShin[5].red, matShin[5].green, matShin[5].blue);
   case 5:
      vpSetMaterial(vpc, VP_MATERIAL4, VP_AMBIENT, VP_BOTH_SIDES, matAmb[4].red, matAmb[4].green, matAmb[4].blue);
      vpSetMaterial(vpc, VP_MATERIAL4, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[4].red, matDiff[4].green, matDiff[4].blue);
      vpSetMaterial(vpc, VP_MATERIAL4, VP_SPECULAR, VP_BOTH_SIDES, matSpec[4].red, matSpec[4].green, matSpec[4].blue);
      vpSetMaterial(vpc, VP_MATERIAL4, VP_SHINYNESS, VP_BOTH_SIDES, matShin[4].red, matShin[4].green, matShin[4].blue);
   case 4:
      vpSetMaterial(vpc, VP_MATERIAL3, VP_AMBIENT, VP_BOTH_SIDES, matAmb[3].red, matAmb[3].green, matAmb[3].blue);
      vpSetMaterial(vpc, VP_MATERIAL3, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[3].red, matDiff[3].green, matDiff[3].blue);
      vpSetMaterial(vpc, VP_MATERIAL3, VP_SPECULAR, VP_BOTH_SIDES, matSpec[3].red, matSpec[3].green, matSpec[3].blue);
      vpSetMaterial(vpc, VP_MATERIAL3, VP_SHINYNESS, VP_BOTH_SIDES, matShin[3].red, matShin[3].green, matShin[3].blue);
   case 3:
      vpSetMaterial(vpc, VP_MATERIAL2, VP_AMBIENT, VP_BOTH_SIDES, matAmb[2].red, matAmb[2].green, matAmb[2].blue);
      vpSetMaterial(vpc, VP_MATERIAL2, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[2].red, matDiff[2].green, matDiff[2].blue);
      vpSetMaterial(vpc, VP_MATERIAL2, VP_SPECULAR, VP_BOTH_SIDES, matSpec[2].red, matSpec[2].green, matSpec[2].blue);
      vpSetMaterial(vpc, VP_MATERIAL2, VP_SHINYNESS, VP_BOTH_SIDES, matShin[2].red, matShin[2].green, matShin[2].blue);
   case 2:
      vpSetMaterial(vpc, VP_MATERIAL1, VP_AMBIENT, VP_BOTH_SIDES, matAmb[1].red, matAmb[1].green, matAmb[1].blue);
      vpSetMaterial(vpc, VP_MATERIAL1, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[1].red, matDiff[1].green, matDiff[1].blue);
      vpSetMaterial(vpc, VP_MATERIAL1, VP_SPECULAR, VP_BOTH_SIDES, matSpec[1].red, matSpec[1].green, matSpec[1].blue);
      vpSetMaterial(vpc, VP_MATERIAL1, VP_SHINYNESS, VP_BOTH_SIDES, matShin[1].red, matShin[1].green, matShin[1].blue);
   case 1:
      vpSetMaterial(vpc, VP_MATERIAL0, VP_AMBIENT, VP_BOTH_SIDES, matAmb[0].red, matAmb[0].green, matAmb[0].blue);
      vpSetMaterial(vpc, VP_MATERIAL0, VP_DIFFUSE, VP_BOTH_SIDES, matDiff[0].red, matDiff[0].green, matDiff[0].blue);
      vpSetMaterial(vpc, VP_MATERIAL0, VP_SPECULAR, VP_BOTH_SIDES, matSpec[0].red, matSpec[0].green, matSpec[0].blue);
      vpSetMaterial(vpc, VP_MATERIAL0, VP_SHINYNESS, VP_BOTH_SIDES, matShin[0].red, matShin[0].green, matShin[0].blue);
      break;
   }
}

void set_lights(){
   
   switch(numLights){
   case 6:
      vpSetLight(vpc, VP_LIGHT5, VP_DIRECTION, lightDir[5].x, lightDir[5].y, lightDir[5].z);
      vpSetLight(vpc, VP_LIGHT5, VP_COLOR, lightColor[5].red, lightColor[5].green, lightColor[5].blue);
      vpEnable(vpc, VP_LIGHT5, 1);   
   case 5:
      vpSetLight(vpc, VP_LIGHT4, VP_DIRECTION, lightDir[4].x, lightDir[4].y, lightDir[4].z);
      vpSetLight(vpc, VP_LIGHT4, VP_COLOR, lightColor[4].red, lightColor[4].green, lightColor[4].blue);      
      vpEnable(vpc, VP_LIGHT4, 1);
   case 4:
      vpSetLight(vpc, VP_LIGHT3, VP_DIRECTION, lightDir[3].x, lightDir[3].y, lightDir[3].z);
      vpSetLight(vpc, VP_LIGHT3, VP_COLOR, lightColor[3].red, lightColor[3].green, lightColor[3].blue);
      vpEnable(vpc, VP_LIGHT3, 1);
   case 3:
      vpSetLight(vpc, VP_LIGHT2, VP_DIRECTION, lightDir[2].x, lightDir[2].y, lightDir[2].z);
      vpSetLight(vpc, VP_LIGHT2, VP_COLOR, lightColor[2].red, lightColor[2].green, lightColor[2].blue);
      vpEnable(vpc, VP_LIGHT2, 1);   
   case 2:
      vpSetLight(vpc, VP_LIGHT1, VP_DIRECTION, lightDir[1].x, lightDir[1].y, lightDir[1].z);
      vpSetLight(vpc, VP_LIGHT1, VP_COLOR, lightColor[1].red, lightColor[1].green, lightColor[1].blue);
      vpEnable(vpc, VP_LIGHT1, 1);
   case 1:
      vpSetLight(vpc, VP_LIGHT0, VP_DIRECTION, lightDir[0].x, lightDir[0].y, lightDir[0].z);
      vpSetLight(vpc, VP_LIGHT0, VP_COLOR, lightColor[0].red, lightColor[0].green, lightColor[0].blue);
      vpEnable(vpc, VP_LIGHT0, 1);
      break;
   }

   switch(numLights){
   case 1:
      vpEnable(vpc, VP_LIGHT1, 0);
   case 2:
      vpEnable(vpc, VP_LIGHT2, 0);
   case 3:
      vpEnable(vpc, VP_LIGHT3, 0);
   case 4:
      vpEnable(vpc, VP_LIGHT4, 0);
   case 5:
      vpEnable(vpc, VP_LIGHT5, 0);
      break;
   }
}



/* ----------------------------------------*
 * Functions which read values from gtk gui* 
 * ----------------------------------------*/
void read_materials(){
   int i;
   GtkAdjustment *adj;
   
   for(i = 0; i < 6; i++){
      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][0]));
      matAmb[i].red = adj->value;
      
      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][1]));
      matAmb[i].green = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][2]));
      matAmb[i].blue = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][0]));
      matDiff[i].red = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][1]));
      matDiff[i].green = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][2]));
      matDiff[i].blue = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][0]));
      matSpec[i].red = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][1]));
      matSpec[i].green = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][2]));
      matSpec[i].blue = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(matShinyScale[i]));
      matShin[i].red = adj->value;

      matShin[i].green = 0.0;
      matShin[i].blue = 0.0;
   }
   
}

void read_lights(){
   int i;
   GtkAdjustment *adj;
   
   for(i = 0; i < 6; i++){
      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][0]));
      lightDir[i].x = adj->value;
      
      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][1]));
      lightDir[i].y = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][2]));
      lightDir[i].z = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][0]));
      lightColor[i].red = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][1]));
      lightColor[i].green = adj->value;

      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][2]));
      lightColor[i].blue = adj->value;
   }
   
}

void read_weights(){
   int i;

   for(i = 0; i < numMaterials; i++){
      weightTable[i][0] = atoi( (char*) gtk_editable_get_chars(GTK_EDITABLE(weightEntry[i][0]), 0, -1));
      weightTable[i][1] = atoi( (char*) gtk_editable_get_chars(GTK_EDITABLE(weightEntry[i][1]), 0, -1));
      fprintf(stderr, "%d - %d\n", weightTable[i][0], weightTable[i][1]);
   }
}

/* ------------------------------------ *
 * Gtk Callbacks which affect rendering *
 * ------------------------------------ */

void rotateX(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_X_AXIS, 10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}

void rotateMinusX(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_X_AXIS, -10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}

void rotateY(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_Y_AXIS, 10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}

void rotateMinusY(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_Y_AXIS, -10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}

void rotateZ(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_Z_AXIS, 10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}

void rotateMinusZ(){
   vpCurrentMatrix(vpc, VP_VIEW);
   vpRotate(vpc, VP_Z_AXIS, -10.0);
   render_volume();
   vpCurrentMatrix(vpc, VP_MODEL);
}
